# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import json
import torch
import torch.nn.functional as F
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
import diffusers.schedulers as schedulers
from peft import LoraConfig
from transformers import (
    Qwen2_5_VLConfig,
    Qwen2_5_VLForConditionalGeneration,
)
from diffusers.schedulers import SchedulerMixin, FlowMatchEulerDiscreteScheduler
from diffusers.training_utils import (
    compute_loss_weighting_for_sd3,
    compute_density_for_timestep_sampling,
)
from diffusers.models import (
    QwenImageTransformer2DModel,
    AutoencoderKLQwenImage,
)
from diffusers.pipelines import (
    QwenImagePipeline,
    QwenImageEditPipeline,
)
from unitorch.models import (
    GenericModel,
    GenericOutputs,
    QuantizationConfig,
    QuantizationMixin,
)
from unitorch.models.peft import GenericPeftModel
from unitorch.models.diffusers import compute_snr
from unitorch.models.diffusers.modeling_qwen_image import (
    _pack_latents,
    _unpack_latents,
)


class GenericQWenImageLoraModel(GenericPeftModel, QuantizationMixin):
    prefix_keys_in_state_dict = {
        # vae weights
        "^encoder.*": "vae.",
        "^decoder.*": "vae.",
        "^post_quant_conv.*": "vae.",
        "^quant_conv.*": "vae.",
    }

    replace_keys_in_state_dict = {
        "\.query\.": ".to_q.",
        "\.key\.": ".to_k.",
        "\.value\.": ".to_v.",
        "\.proj_attn\.": ".to_out.0.",
    }

    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        image_config_path: Optional[str] = None,
        quant_config_path: Optional[str] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        snr_gamma: Optional[float] = 5.0,
        lora_r: Optional[int] = 16,
        lora_alpha: Optional[int] = 32,
        lora_dropout: Optional[float] = 0.05,
        fan_in_fan_out: Optional[bool] = True,
        target_modules: Optional[Union[List[str], str]] = [
            "to_q",
            "to_k",
            "to_v",
            "q_proj",
            "k_proj",
            "v_proj",
            "add_q_proj",
            "add_k_proj",
            "add_v_proj",
        ],
        enable_text_adapter: Optional[bool] = True,
        enable_transformer_adapter: Optional[bool] = True,
        seed: Optional[int] = 1123,
    ):
        super().__init__()
        self.seed = seed
        self.num_train_timesteps = num_train_timesteps
        self.num_infer_timesteps = num_infer_timesteps
        self.snr_gamma = snr_gamma

        config_dict = json.load(open(config_path))
        self.transformer = QwenImageTransformer2DModel.from_config(config_dict).to(
            torch.bfloat16
        )

        text_config = Qwen2_5_VLConfig.from_json_file(text_config_path)
        self.text = Qwen2_5_VLForConditionalGeneration(text_config).to(torch.bfloat16)

        vae_config_dict = json.load(open(vae_config_path))
        self.vae = AutoencoderKLQwenImage.from_config(vae_config_dict).to(
            torch.bfloat16
        )

        scheduler_config_dict = json.load(open(scheduler_config_path))
        scheduler_class_name = scheduler_config_dict.get(
            "_class_name", "FlowMatchEulerDiscreteScheduler"
        )
        assert hasattr(schedulers, scheduler_class_name)
        scheduler_class = getattr(schedulers, scheduler_class_name)
        assert issubclass(scheduler_class, SchedulerMixin)
        scheduler_config_dict["num_train_timesteps"] = num_train_timesteps
        self.scheduler = scheduler_class.from_config(scheduler_config_dict)

        for param in self.vae.parameters():
            param.requires_grad = False

        for param in self.text.parameters():
            param.requires_grad = False

        for param in self.transformer.parameters():
            param.requires_grad = False

        if quant_config_path is not None:
            self.quant_config = QuantizationConfig.from_json_file(quant_config_path)
            self.quantize(
                self.quant_config, ignore_modules=["lm_head", "transformer", "vae"]
            )

        lora_config = LoraConfig(
            r=lora_r,
            lora_alpha=lora_alpha,
            init_lora_weights="gaussian",
            lora_dropout=lora_dropout,
            fan_in_fan_out=fan_in_fan_out,
            target_modules=target_modules,
        )
        if enable_text_adapter:
            self.text.add_adapter(lora_config)
        if enable_transformer_adapter:
            self.transformer.add_adapter(lora_config)

    def get_sigmas(self, timesteps, n_dim=4, dtype=torch.float32):
        sigmas = self.scheduler.sigmas.to(device=self.device, dtype=dtype)
        schedule_timesteps = self.scheduler.timesteps.to(self.device)
        timesteps = timesteps.to(self.device)
        step_indices = [(schedule_timesteps == t).nonzero().item() for t in timesteps]

        sigma = sigmas[step_indices].flatten()
        while len(sigma.shape) < n_dim:
            sigma = sigma.unsqueeze(-1)
        return sigma

    def _extract_masked_hidden(self, hidden_states: torch.Tensor, mask: torch.Tensor):
        bool_mask = mask.bool()
        valid_lengths = bool_mask.sum(dim=1)
        selected = hidden_states[bool_mask]
        split_result = torch.split(selected, valid_lengths.tolist(), dim=0)

        return split_result

    def get_prompt_outputs(
        self,
        input_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        pixel_values: Optional[torch.Tensor] = None,
        image_grid_thw: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        prompt_start_index: Optional[int] = None,
        enable_cpu_offload: Optional[bool] = False,
        cpu_offload_device: Optional[str] = "cpu",
    ):
        if enable_cpu_offload:
            self.text = self.text.to(cpu_offload_device)
            input_ids = input_ids.to(cpu_offload_device)
            negative_input_ids = negative_input_ids.to(cpu_offload_device)
            if pixel_values is not None:
                pixel_values = pixel_values.to(cpu_offload_device)
            if image_grid_thw is not None:
                image_grid_thw = image_grid_thw.to(cpu_offload_device)
            if attention_mask is not None:
                attention_mask = attention_mask.to(cpu_offload_device)
            if negative_attention_mask is not None:
                negative_attention_mask = negative_attention_mask.to(cpu_offload_device)

        assert prompt_start_index is not None, "prompt_start_index must be provided"

        if pixel_values is not None and image_grid_thw is not None:
            image_grid_thw = image_grid_thw.view(-1, image_grid_thw.size(-1))
            pixel_values = pixel_values.view(-1, pixel_values.size(-1))

        prompt_outputs = self.text(
            input_ids=input_ids,
            attention_mask=attention_mask,
            pixel_values=pixel_values,
            image_grid_thw=image_grid_thw,
            output_hidden_states=True,
        )

        hidden_states = prompt_outputs.hidden_states[-1]
        split_hidden_states = self._extract_masked_hidden(hidden_states, attention_mask)
        split_hidden_states = [e[prompt_start_index:] for e in split_hidden_states]
        attn_mask_list = [
            torch.ones(e.size(0), dtype=torch.long, device=e.device)
            for e in split_hidden_states
        ]
        max_seq_len = max([e.size(0) for e in split_hidden_states])
        prompt_embeds = torch.stack(
            [
                torch.cat([u, u.new_zeros(max_seq_len - u.size(0), u.size(1))])
                for u in split_hidden_states
            ]
        )
        prompt_embeds_mask = torch.stack(
            [
                torch.cat([u, u.new_zeros(max_seq_len - u.size(0))])
                for u in attn_mask_list
            ]
        )

        negative_prompt_outputs = self.text(
            negative_input_ids,
            negative_attention_mask,
            output_hidden_states=True,
        )
        negative_hidden_states = negative_prompt_outputs.hidden_states[-1]
        negative_split_hidden_states = self._extract_masked_hidden(
            negative_hidden_states, negative_attention_mask
        )
        negative_split_hidden_states = [
            e[prompt_start_index:] for e in negative_split_hidden_states
        ]
        negative_attn_mask_list = [
            torch.ones(e.size(0), dtype=torch.long, device=e.device)
            for e in negative_split_hidden_states
        ]
        negative_max_seq_len = max([e.size(0) for e in negative_split_hidden_states])
        negative_prompt_embeds = torch.stack(
            [
                torch.cat(
                    [
                        u,
                        u.new_zeros(negative_max_seq_len - u.size(0), u.size(1)),
                    ]
                )
                for u in negative_split_hidden_states
            ]
        )
        negative_prompt_embeds_mask = torch.stack(
            [
                torch.cat(
                    [
                        u,
                        u.new_zeros(negative_max_seq_len - u.size(0)),
                    ]
                )
                for u in negative_attn_mask_list
            ]
        )

        if enable_cpu_offload:
            self.text = self.text.to("cpu")

        return GenericOutputs(
            prompt_embeds=prompt_embeds,
            prompt_embeds_mask=prompt_embeds_mask,
            negative_prompt_embeds=negative_prompt_embeds,
            negative_prompt_embeds_mask=negative_prompt_embeds_mask,
        )


class QWenImageLoraForText2ImageGeneration(GenericQWenImageLoraModel):
    modules_to_save_checkpoints = ["lora"]

    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        snr_gamma: Optional[float] = 5.0,
        lora_r: Optional[int] = 16,
        lora_alpha: Optional[int] = 32,
        lora_dropout: Optional[float] = 0.05,
        fan_in_fan_out: Optional[bool] = True,
        target_modules: Optional[Union[List[str], str]] = [
            "to_q",
            "to_k",
            "to_v",
            "q_proj",
            "k_proj",
            "v_proj",
            "add_q_proj",
            "add_k_proj",
            "add_v_proj",
        ],
        enable_text_adapter: Optional[bool] = True,
        enable_transformer_adapter: Optional[bool] = True,
        seed: Optional[int] = 1123,
        gradient_checkpointing: Optional[bool] = True,
        guidance_scale: Optional[float] = 1.0,
        prompt_start_index: Optional[int] = 34,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            snr_gamma=snr_gamma,
            lora_r=lora_r,
            lora_alpha=lora_alpha,
            lora_dropout=lora_dropout,
            fan_in_fan_out=fan_in_fan_out,
            target_modules=target_modules,
            enable_text_adapter=enable_text_adapter,
            enable_transformer_adapter=enable_transformer_adapter,
            seed=seed,
        )

        if gradient_checkpointing:
            self.transformer.enable_gradient_checkpointing()

        self.pipeline = QwenImagePipeline(
            vae=self.vae,
            text_encoder=self.text,
            transformer=self.transformer,
            scheduler=self.scheduler,
            tokenizer=None,
        )
        self.pipeline.set_progress_bar_config(disable=True)
        self.guidance_scale = guidance_scale
        self.prompt_start_index = prompt_start_index

    def forward(
        self,
        input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
    ):
        prompt_outputs = self.text(
            input_ids=input_ids,
            attention_mask=attention_mask,
            output_hidden_states=True,
        )

        hidden_states = prompt_outputs.hidden_states[-1]
        split_hidden_states = self._extract_masked_hidden(hidden_states, attention_mask)
        split_hidden_states = [
            e[self.prompt_start_index :] for e in split_hidden_states
        ]
        attn_mask_list = [
            torch.ones(e.size(0), dtype=torch.long, device=e.device)
            for e in split_hidden_states
        ]
        max_seq_len = max([e.size(0) for e in split_hidden_states])
        prompt_embeds = torch.stack(
            [
                torch.cat([u, u.new_zeros(max_seq_len - u.size(0), u.size(1))])
                for u in split_hidden_states
            ]
        )
        prompt_embeds_mask = torch.stack(
            [
                torch.cat([u, u.new_zeros(max_seq_len - u.size(0))])
                for u in attn_mask_list
            ]
        )
        if pixel_values.ndim == 4:
            pixel_values = pixel_values.unsqueeze(2)

        latents = self.vae.encode(pixel_values).latent_dist.sample()
        latents_mean = (
            torch.tensor(self.vae.config.latents_mean).view(
                1, self.vae.config.z_dim, 1, 1, 1
            )
        ).to(pixel_values.device)
        latents_std = 1.0 / torch.tensor(self.vae.config.latents_std).view(
            1, self.vae.config.z_dim, 1, 1, 1
        ).to(pixel_values.device)
        latents = (latents - latents_mean) * latents_std

        noise = torch.randn(latents.shape).to(latents.device)

        u = compute_density_for_timestep_sampling(
            weighting_scheme="none",
            batch_size=latents.shape[0],
            logit_mean=0.0,
            logit_std=1.0,
            mode_scale=1.29,
        )
        indices = (u * self.scheduler.config.num_train_timesteps).long()
        timesteps = self.scheduler.timesteps[indices].to(device=self.device)

        sigmas = self.get_sigmas(timesteps, n_dim=latents.ndim, dtype=latents.dtype)
        noise_latents = (1.0 - sigmas) * latents + sigmas * noise

        noise_latents = _pack_latents(
            noise_latents,
            batch_size=latents.shape[0],
            num_channels_latents=latents.shape[1],
            height=latents.shape[-2],
            width=latents.shape[-1],
        )

        vae_scale_factor = 2 ** len(self.vae.temperal_downsample)
        batch_size, width, height = (
            pixel_values.size(0),
            pixel_values.size(-1),
            pixel_values.size(-2),
        )
        img_shapes = [
            [
                (
                    1,
                    height // vae_scale_factor // 2,
                    width // vae_scale_factor // 2,
                )
            ]
        ] * batch_size

        if self.transformer.config.guidance_embeds and self.guidance_scale is not None:
            guidance = torch.full(
                [1], self.guidance_scale, device=self.device, dtype=torch.float32
            )
            guidance = guidance.expand(latents.shape[0])
        else:
            guidance = None

        outputs = self.transformer(
            hidden_states=noise_latents,
            timestep=timesteps / 1000,
            guidance=guidance,
            encoder_hidden_states=prompt_embeds,
            encoder_hidden_states_mask=prompt_embeds_mask,
            img_shapes=img_shapes,
            txt_seq_lens=prompt_embeds_mask.sum(dim=1).tolist(),
            return_dict=False,
        )[0]

        outputs = _unpack_latents(
            outputs,
            height=latents.shape[-2] * vae_scale_factor,
            width=latents.shape[-1] * vae_scale_factor,
            vae_scale_factor=vae_scale_factor,
        )

        weighting = compute_loss_weighting_for_sd3(
            weighting_scheme="none", sigmas=sigmas
        )
        target = noise - latents
        loss = torch.mean(
            (weighting.float() * (outputs.float() - target.float()) ** 2).reshape(
                target.shape[0], -1
            ),
            1,
        )
        loss = loss.mean()
        return loss

    def generate(
        self,
        input_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        height: Optional[int] = 1024,
        width: Optional[int] = 1024,
        guidance_scale: Optional[float] = 1.0,
        true_cfg_scale: Optional[float] = 4.0,
    ):
        outputs = self.get_prompt_outputs(
            input_ids=input_ids,
            negative_input_ids=negative_input_ids,
            attention_mask=attention_mask,
            negative_attention_mask=negative_attention_mask,
            prompt_start_index=self.prompt_start_index,
        )

        images = self.pipeline(
            prompt_embeds=outputs.prompt_embeds,
            negative_prompt_embeds=outputs.negative_prompt_embeds,
            prompt_embeds_mask=outputs.prompt_embeds_mask,
            negative_prompt_embeds_mask=outputs.negative_prompt_embeds_mask,
            generator=torch.Generator(device=self.pipeline.device).manual_seed(
                self.seed
            ),
            num_inference_steps=self.num_infer_timesteps,
            height=height,
            width=width,
            guidance_scale=guidance_scale,
            true_cfg_scale=true_cfg_scale,
            output_type="np.array",
        ).images

        return GenericOutputs(images=torch.from_numpy(images))


class QWenImageLoraForImageEditing(GenericQWenImageLoraModel):
    modules_to_save_checkpoints = ["lora"]

    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        snr_gamma: Optional[float] = 5.0,
        lora_r: Optional[int] = 16,
        lora_alpha: Optional[int] = 32,
        lora_dropout: Optional[float] = 0.05,
        fan_in_fan_out: Optional[bool] = True,
        target_modules: Optional[Union[List[str], str]] = [
            "to_q",
            "to_k",
            "to_v",
            "q_proj",
            "k_proj",
            "v_proj",
            "add_q_proj",
            "add_k_proj",
            "add_v_proj",
        ],
        enable_text_adapter: Optional[bool] = True,
        enable_transformer_adapter: Optional[bool] = True,
        seed: Optional[int] = 1123,
        gradient_checkpointing: Optional[bool] = True,
        guidance_scale: Optional[float] = 1.0,
        prompt_start_index: Optional[int] = 64,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            snr_gamma=snr_gamma,
            lora_r=lora_r,
            lora_alpha=lora_alpha,
            lora_dropout=lora_dropout,
            fan_in_fan_out=fan_in_fan_out,
            target_modules=target_modules,
            enable_text_adapter=enable_text_adapter,
            enable_transformer_adapter=enable_transformer_adapter,
            seed=seed,
        )
        if gradient_checkpointing:
            self.transformer.enable_gradient_checkpointing()

        self.pipeline = QwenImageEditPipeline(
            vae=self.vae,
            text_encoder=self.text,
            transformer=self.transformer,
            scheduler=self.scheduler,
            tokenizer=None,
            processor=None,
        )
        self.pipeline.set_progress_bar_config(disable=True)
        self.guidance_scale = guidance_scale
        self.prompt_start_index = prompt_start_index

    def forward(
        self,
        input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        refer_pixel_values: torch.Tensor,
        refer_image_grid_thw: torch.Tensor,
        refer_vae_pixel_values: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
    ):
        refer_image_grid_thw = refer_image_grid_thw.view(
            -1, refer_image_grid_thw.size(-1)
        )
        refer_pixel_values = refer_pixel_values.view(-1, refer_pixel_values.size(-1))
        prompt_outputs = self.text(
            input_ids=input_ids,
            attention_mask=attention_mask,
            pixel_values=refer_pixel_values,
            image_grid_thw=refer_image_grid_thw,
            output_hidden_states=True,
        )

        hidden_states = prompt_outputs.hidden_states[-1]
        split_hidden_states = self._extract_masked_hidden(hidden_states, attention_mask)
        split_hidden_states = [
            e[self.prompt_start_index :] for e in split_hidden_states
        ]
        attn_mask_list = [
            torch.ones(e.size(0), dtype=torch.long, device=e.device)
            for e in split_hidden_states
        ]
        max_seq_len = max([e.size(0) for e in split_hidden_states])
        prompt_embeds = torch.stack(
            [
                torch.cat([u, u.new_zeros(max_seq_len - u.size(0), u.size(1))])
                for u in split_hidden_states
            ]
        )
        prompt_embeds_mask = torch.stack(
            [
                torch.cat([u, u.new_zeros(max_seq_len - u.size(0))])
                for u in attn_mask_list
            ]
        )

        if pixel_values.ndim == 4:
            pixel_values = pixel_values.unsqueeze(2)
        if refer_vae_pixel_values.ndim == 4:
            refer_vae_pixel_values = refer_vae_pixel_values.unsqueeze(2)

        latents = self.vae.encode(pixel_values).latent_dist.sample()
        refer_latents = self.vae.encode(refer_vae_pixel_values).latent_dist.mode()
        latents_mean = (
            torch.tensor(self.vae.config.latents_mean).view(
                1, self.vae.config.z_dim, 1, 1, 1
            )
        ).to(pixel_values.device)
        latents_std = 1.0 / torch.tensor(self.vae.config.latents_std).view(
            1, self.vae.config.z_dim, 1, 1, 1
        ).to(pixel_values.device)
        latents = (latents - latents_mean) * latents_std
        refer_latents = (refer_latents - latents_mean) * latents_std

        noise = torch.randn(latents.shape).to(latents.device)

        u = compute_density_for_timestep_sampling(
            weighting_scheme="none",
            batch_size=latents.shape[0],
            logit_mean=0.0,
            logit_std=1.0,
            mode_scale=1.29,
        )
        indices = (u * self.scheduler.config.num_train_timesteps).long()
        timesteps = self.scheduler.timesteps[indices].to(device=self.device)

        sigmas = self.get_sigmas(timesteps, n_dim=latents.ndim, dtype=latents.dtype)
        noise_latents = (1.0 - sigmas) * latents + sigmas * noise

        noise_latents = _pack_latents(
            noise_latents,
            batch_size=latents.shape[0],
            num_channels_latents=latents.shape[1],
            height=latents.shape[-2],
            width=latents.shape[-1],
        )
        refer_latents = _pack_latents(
            refer_latents,
            batch_size=refer_latents.shape[0],
            num_channels_latents=refer_latents.shape[1],
            height=refer_latents.shape[-2],
            width=refer_latents.shape[-1],
        )
        latent_model_input = torch.cat([noise_latents, refer_latents], dim=1)

        vae_scale_factor = 2 ** len(self.vae.temperal_downsample)
        batch_size, width, height = (
            pixel_values.size(0),
            pixel_values.size(-1),
            pixel_values.size(-2),
        )
        refer_width, refer_height = (
            refer_vae_pixel_values.size(-1),
            refer_vae_pixel_values.size(-2),
        )
        img_shapes = [
            [
                (
                    1,
                    height // vae_scale_factor // 2,
                    width // vae_scale_factor // 2,
                ),
                (
                    1,
                    refer_height // vae_scale_factor // 2,
                    refer_width // vae_scale_factor // 2,
                ),
            ]
        ] * batch_size

        if self.transformer.config.guidance_embeds and self.guidance_scale is not None:
            guidance = torch.full(
                [1], self.guidance_scale, device=self.device, dtype=torch.float32
            )
            guidance = guidance.expand(latents.shape[0])
        else:
            guidance = None

        outputs = self.transformer(
            hidden_states=latent_model_input,
            timestep=timesteps / 1000,
            guidance=guidance,
            encoder_hidden_states=prompt_embeds,
            encoder_hidden_states_mask=prompt_embeds_mask,
            img_shapes=img_shapes,
            txt_seq_lens=prompt_embeds_mask.sum(dim=1).tolist(),
            return_dict=False,
        )[0]
        outputs = outputs[:, : noise_latents.shape[1]]

        outputs = _unpack_latents(
            outputs,
            height=latents.shape[-2] * vae_scale_factor,
            width=latents.shape[-1] * vae_scale_factor,
            vae_scale_factor=vae_scale_factor,
        )

        weighting = compute_loss_weighting_for_sd3(
            weighting_scheme="none", sigmas=sigmas
        )
        target = noise - latents
        loss = torch.mean(
            (weighting.float() * (outputs.float() - target.float()) ** 2).reshape(
                target.shape[0], -1
            ),
            1,
        )
        loss = loss.mean()
        return loss

    def generate(
        self,
        input_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        refer_pixel_values: torch.Tensor,
        refer_image_grid_thw: torch.Tensor,
        refer_vae_pixel_values: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        height: Optional[int] = 1024,
        width: Optional[int] = 1024,
        guidance_scale: Optional[float] = 1.0,
        true_cfg_scale: Optional[float] = 4.0,
    ):
        outputs = self.get_prompt_outputs(
            input_ids=input_ids,
            negative_input_ids=negative_input_ids,
            pixel_values=refer_pixel_values,
            image_grid_thw=refer_image_grid_thw,
            attention_mask=attention_mask,
            negative_attention_mask=negative_attention_mask,
            prompt_start_index=self.prompt_start_index,
        )

        images = self.pipeline(
            images=refer_vae_pixel_values,
            prompt_embeds=outputs.prompt_embeds,
            negative_prompt_embeds=outputs.negative_prompt_embeds,
            prompt_embeds_mask=outputs.prompt_embeds_mask,
            negative_prompt_embeds_mask=outputs.negative_prompt_embeds_mask,
            generator=torch.Generator(device=self.pipeline.device).manual_seed(
                self.seed
            ),
            num_inference_steps=self.num_infer_timesteps,
            height=height,
            width=width,
            guidance_scale=guidance_scale,
            true_cfg_scale=true_cfg_scale,
            output_type="np.array",
        ).images

        return GenericOutputs(images=torch.from_numpy(images))



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import re
import json
import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from peft import LoraConfig, PeftModelForCausalLM
from transformers.models.qwen3 import Qwen3Config, Qwen3ForCausalLM
from unitorch.models import GenericModel, GenericOutputs
from unitorch.models.peft import PeftModelForSequenceClassification, GenericPeftModel


class QWen3LoraForGeneration(GenericPeftModel):
    prefix_keys_in_state_dict = {"^(?!model\.model\.).*": "model."}
    replace_keys_in_state_dict = {
        "q_proj.weight": "q_proj.base_layer.weight",
        "q_proj.bias": "q_proj.base_layer.bias",
        "v_proj.weight": "v_proj.base_layer.weight",
        "v_proj.bias": "v_proj.base_layer.bias",
    }

    def __init__(
        self,
        config_path: str,
        lora_r: Optional[int] = 16,
        lora_alpha: Optional[int] = 32,
        lora_dropout: Optional[float] = 0.05,
        fan_in_fan_out: Optional[bool] = True,
        target_modules: Optional[Union[List[str], str]] = ["q_proj", "v_proj"],
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Bloom Lora model for text generation tasks.

        Args:
            config_path (str): Path to the model configuration file.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__()
        self.config = Qwen3Config.from_json_file(config_path)
        self.config.gradient_checkpointing = gradient_checkpointing
        self.peft_config = LoraConfig(
            r=lora_r,
            lora_alpha=lora_alpha,
            lora_dropout=lora_dropout,
            fan_in_fan_out=fan_in_fan_out,
            target_modules=target_modules,
        )
        self.model = Qwen3ForCausalLM(self.config)
        self.model.add_adapter(self.peft_config)
        self.init_weights()

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
    ):
        """
        Forward pass of the generation model.

        Args:
            input_ids (torch.Tensor, optional): Input tensor of shape (batch_size, sequence_length). Defaults to None.
            attention_mask (torch.Tensor, optional): Attention mask tensor of shape (batch_size, sequence_length). Defaults to None.
            position_ids (torch.Tensor, optional): Position IDs tensor of shape (batch_size, sequence_length). Defaults to None.

        Returns:
            torch Output logits.Tensor: tensor of shape (batch_size, sequence_length, vocab_size).
        """
        outputs = self.model(
            input_ids=input_ids,
            attention_mask=attention_mask,
            return_dict=True,
        )
        logits = outputs.logits
        return logits

    @torch.no_grad()
    def generate(
        self,
        input_ids: torch.Tensor,
        num_beams: Optional[int] = 5,
        decoder_start_token_id: Optional[int] = 151643,
        decoder_end_token_id: Optional[Union[int, List[int]]] = 151645,
        decoder_pad_token_id: Optional[int] = 151643,
        num_return_sequences: Optional[int] = 1,
        min_gen_seq_length: Optional[int] = 0,
        max_gen_seq_length: Optional[int] = 512,
        repetition_penalty: Optional[float] = 1.0,
        no_repeat_ngram_size: Optional[int] = 0,
        early_stopping: Optional[bool] = True,
        length_penalty: Optional[float] = 1.0,
        num_beam_groups: Optional[int] = 1,
        diversity_penalty: Optional[float] = 0.0,
        do_sample: Optional[bool] = False,
        temperature: Optional[float] = 1.0,
        top_k: Optional[int] = 50,
        top_p: Optional[float] = 1.0,
    ):
        """
        Generate text using the generation model.

        Args:
            input_ids: Input tensor of shape (batch_size, sequence_length).
            num_beams (int, optional): Number of beams for beam search. Defaults to 5.
            decoder_start_token_id (int, optional): The ID of the decoder start token. Defaults to 2.
            decoder_end_token_id (int or List[int], optional): The ID(s) of the decoder end token(s). Defaults to 2.
            num_return_sequences (int, optional): Number of generated sequences to return. Defaults to 1.
            min_gen_seq_length (int, optional): Minimum length of generated sequences. Defaults to 0.
            max_gen_seq_length (int, optional): Maximum length of generated sequences. Defaults to 48.
            repetition_penalty (float, optional): Penalty for repeated tokens. Defaults to 1.0.
            no_repeat_ngram_size (int, optional): Size of n-grams to avoid repeating. Defaults to 0.
            early_stopping (bool, optional): Whether to stop generation early. Defaults to True.
            length_penalty (float, optional): Penalty for longer sequences. Defaults to 1.0.
            num_beam_groups (int, optional): Number of beam groups for diverse beam search. Defaults to 1.
            diversity_penalty (float, optional): Penalty for diverse sequences in diverse beam search. Defaults to 0.0.
            do_sample (bool, optional): Whether to use sampling for generation. Defaults to False.
            temperature (float, optional): Sampling temperature. Defaults to 1.0.
            top_k (int, optional): Top-k value for sampling. Defaults to 50.
            top_p (float, optional): Top-p value for sampling. Defaults to 1.0.

        Returns:
            GenericOutputs: Generated sequences and their scores.
        """
        input_seq_length = input_ids.size(1)
        outputs = self.model.generate(
            input_ids=input_ids,
            max_length=max_gen_seq_length + input_seq_length,
            min_length=min_gen_seq_length + input_seq_length,
            num_beams=num_beams,
            do_sample=do_sample,
            no_repeat_ngram_size=no_repeat_ngram_size,
            early_stopping=early_stopping,
            length_penalty=length_penalty,
            repetition_penalty=repetition_penalty,
            num_return_sequences=num_return_sequences,
            bos_token_id=decoder_start_token_id,
            eos_token_id=decoder_end_token_id,
            pad_token_id=decoder_pad_token_id,
            num_beam_groups=num_beam_groups,
            diversity_penalty=diversity_penalty,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            return_dict_in_generate=True,
            output_scores=True,
        )

        sequences = outputs.sequences.reshape(
            -1, num_return_sequences, outputs.sequences.size(-1)
        )
        outputs.sequences = (
            torch.zeros(sequences.size(0), num_return_sequences, max_gen_seq_length).to(
                device=sequences.device
            )
            + decoder_start_token_id
        )
        outputs.sequences[:, :, : sequences.size(-1) - input_seq_length].copy_(
            sequences[:, :, input_seq_length : sequences.size(-1)]
        )

        if num_return_sequences == 1:
            outputs.sequences = outputs.sequences.reshape(-1, max_gen_seq_length)

        return GenericOutputs(
            sequences=outputs.sequences.long(),
            sequences_scores=outputs.sequences_scores,
        )


class QWen3DPOLoraForGeneration(GenericPeftModel):
    prefix_keys_in_state_dict = {"^(?!model\.model\.).*": "model."}
    replace_keys_in_state_dict = {
        "q_proj.weight": "q_proj.base_layer.weight",
        "q_proj.bias": "q_proj.base_layer.bias",
        "v_proj.weight": "v_proj.base_layer.weight",
        "v_proj.bias": "v_proj.base_layer.bias",
    }

    def __init__(
        self,
        config_path: str,
        lora_r: Optional[int] = 16,
        lora_alpha: Optional[int] = 32,
        lora_dropout: Optional[float] = 0.05,
        fan_in_fan_out: Optional[bool] = True,
        target_modules: Optional[Union[List[str], str]] = ["q_proj", "v_proj"],
        gradient_checkpointing: Optional[bool] = False,
        dpo_beta: Optional[float] = 0.1,
    ):
        """
        Bloom Lora model for text generation tasks.

        Args:
            config_path (str): Path to the model configuration file.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__()
        self.config = Qwen3Config.from_json_file(config_path)
        self.config.gradient_checkpointing = gradient_checkpointing
        self.peft_config = LoraConfig(
            r=lora_r,
            lora_alpha=lora_alpha,
            lora_dropout=lora_dropout,
            fan_in_fan_out=fan_in_fan_out,
            target_modules=target_modules,
        )
        self.model = Qwen3ForCausalLM(self.config)
        self.model.add_adapter(self.peft_config)
        self.init_weights()
        self.dpo_beta = dpo_beta

    def forward(
        self,
        input_ids: torch.Tensor,
        win_input_ids: torch.Tensor,
        lose_input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        win_attention_mask: Optional[torch.Tensor] = None,
        lose_attention_mask: Optional[torch.Tensor] = None,
    ):
        """
        Forward pass of the generation model.

        Args:
            input_ids (torch.Tensor, optional): Input tensor of shape (batch_size, sequence_length). Defaults to None.
            attention_mask (torch.Tensor, optional): Attention mask tensor of shape (batch_size, sequence_length). Defaults to None.
            position_ids (torch.Tensor, optional): Position IDs tensor of shape (batch_size, sequence_length). Defaults to None.

        Returns:
            torch Output logits.Tensor: tensor of shape (batch_size, sequence_length, vocab_size).
        """
        win_input_ids = torch.cat([input_ids, win_input_ids], dim=1)
        lose_input_ids = torch.cat([input_ids, lose_input_ids], dim=1)
        if attention_mask is not None and win_attention_mask is not None:
            win_attention_mask = torch.cat([attention_mask, win_attention_mask], dim=1)
        if attention_mask is not None and lose_attention_mask is not None:
            lose_attention_mask = torch.cat(
                [attention_mask, lose_attention_mask], dim=1
            )
        win_outputs = self.model(
            input_ids=win_input_ids,
            attention_mask=win_attention_mask,
            return_dict=True,
        )
        lose_outputs = self.model(
            input_ids=lose_input_ids,
            attention_mask=lose_attention_mask,
            return_dict=True,
        )
        input_seq_length = input_ids.size(1)
        win_logits = win_outputs.logits[:, input_seq_length - 1 : -1, :]
        lose_logits = lose_outputs.logits[:, input_seq_length - 1 : -1, :]
        win_labels = win_input_ids[:, input_seq_length:]
        lose_labels = lose_input_ids[:, input_seq_length:]
        win_labels_mask = win_attention_mask[:, input_seq_length:]
        lose_labels_mask = lose_attention_mask[:, input_seq_length:]
        win_nll_loss = F.cross_entropy(
            win_logits.reshape(-1, win_logits.size(-1)),
            win_labels.reshape(-1),
            reduction="none",
        ).reshape(win_labels.size(0), -1)
        win_logprobs = -win_nll_loss * win_labels_mask
        lose_nll_loss = F.cross_entropy(
            lose_logits.reshape(-1, lose_logits.size(-1)),
            lose_labels.reshape(-1),
            reduction="none",
        ).reshape(lose_labels.size(0), -1)
        lose_logprobs = -lose_nll_loss * lose_labels_mask

        with torch.no_grad():
            self.model.disable_adapters()
            ref_win_outputs = self.model(
                input_ids=win_input_ids,
                attention_mask=win_attention_mask,
                return_dict=True,
            )
            ref_lose_outputs = self.model(
                input_ids=lose_input_ids,
                attention_mask=lose_attention_mask,
                return_dict=True,
            )
            ref_win_logits = ref_win_outputs.logits[:, input_seq_length - 1 : -1, :]
            ref_lose_logits = ref_lose_outputs.logits[:, input_seq_length - 1 : -1, :]
            ref_win_nll_loss = F.cross_entropy(
                ref_win_logits.reshape(-1, ref_win_logits.size(-1)),
                win_labels.reshape(-1),
                reduction="none",
            ).reshape(win_labels.size(0), -1)
            ref_win_logprobs = -ref_win_nll_loss * win_labels_mask
            ref_lose_nll_loss = F.cross_entropy(
                ref_lose_logits.reshape(-1, ref_lose_logits.size(-1)),
                lose_labels.reshape(-1),
                reduction="none",
            ).reshape(lose_labels.size(0), -1)
            ref_lose_logprobs = -ref_lose_nll_loss * lose_labels_mask
            self.model.enable_adapters()

        logratios = win_logprobs - lose_logprobs
        ref_logratios = ref_win_logprobs - ref_lose_logprobs
        logits = logratios - ref_logratios
        loss = -F.logsigmoid(self.dpo_beta * logits).mean()
        return loss

    @torch.no_grad()
    def generate(
        self,
        input_ids: torch.Tensor,
        num_beams: Optional[int] = 5,
        decoder_start_token_id: Optional[int] = 151643,
        decoder_end_token_id: Optional[Union[int, List[int]]] = 151645,
        decoder_pad_token_id: Optional[int] = 151643,
        num_return_sequences: Optional[int] = 1,
        min_gen_seq_length: Optional[int] = 0,
        max_gen_seq_length: Optional[int] = 512,
        repetition_penalty: Optional[float] = 1.0,
        no_repeat_ngram_size: Optional[int] = 0,
        early_stopping: Optional[bool] = True,
        length_penalty: Optional[float] = 1.0,
        num_beam_groups: Optional[int] = 1,
        diversity_penalty: Optional[float] = 0.0,
        do_sample: Optional[bool] = False,
        temperature: Optional[float] = 1.0,
        top_k: Optional[int] = 50,
        top_p: Optional[float] = 1.0,
    ):
        """
        Generate text using the generation model.

        Args:
            input_ids: Input tensor of shape (batch_size, sequence_length).
            num_beams (int, optional): Number of beams for beam search. Defaults to 5.
            decoder_start_token_id (int, optional): The ID of the decoder start token. Defaults to 2.
            decoder_end_token_id (int or List[int], optional): The ID(s) of the decoder end token(s). Defaults to 2.
            num_return_sequences (int, optional): Number of generated sequences to return. Defaults to 1.
            min_gen_seq_length (int, optional): Minimum length of generated sequences. Defaults to 0.
            max_gen_seq_length (int, optional): Maximum length of generated sequences. Defaults to 48.
            repetition_penalty (float, optional): Penalty for repeated tokens. Defaults to 1.0.
            no_repeat_ngram_size (int, optional): Size of n-grams to avoid repeating. Defaults to 0.
            early_stopping (bool, optional): Whether to stop generation early. Defaults to True.
            length_penalty (float, optional): Penalty for longer sequences. Defaults to 1.0.
            num_beam_groups (int, optional): Number of beam groups for diverse beam search. Defaults to 1.
            diversity_penalty (float, optional): Penalty for diverse sequences in diverse beam search. Defaults to 0.0.
            do_sample (bool, optional): Whether to use sampling for generation. Defaults to False.
            temperature (float, optional): Sampling temperature. Defaults to 1.0.
            top_k (int, optional): Top-k value for sampling. Defaults to 50.
            top_p (float, optional): Top-p value for sampling. Defaults to 1.0.

        Returns:
            GenericOutputs: Generated sequences and their scores.
        """
        input_seq_length = input_ids.size(1)
        outputs = self.model.generate(
            input_ids=input_ids,
            max_length=max_gen_seq_length + input_seq_length,
            min_length=min_gen_seq_length + input_seq_length,
            num_beams=num_beams,
            do_sample=do_sample,
            no_repeat_ngram_size=no_repeat_ngram_size,
            early_stopping=early_stopping,
            length_penalty=length_penalty,
            repetition_penalty=repetition_penalty,
            num_return_sequences=num_return_sequences,
            bos_token_id=decoder_start_token_id,
            eos_token_id=decoder_end_token_id,
            pad_token_id=decoder_pad_token_id,
            num_beam_groups=num_beam_groups,
            diversity_penalty=diversity_penalty,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            return_dict_in_generate=True,
            output_scores=True,
        )

        sequences = outputs.sequences.reshape(
            -1, num_return_sequences, outputs.sequences.size(-1)
        )
        outputs.sequences = (
            torch.zeros(sequences.size(0), num_return_sequences, max_gen_seq_length).to(
                device=sequences.device
            )
            + decoder_start_token_id
        )
        outputs.sequences[:, :, : sequences.size(-1) - input_seq_length].copy_(
            sequences[:, :, input_seq_length : sequences.size(-1)]
        )

        if num_return_sequences == 1:
            outputs.sequences = outputs.sequences.reshape(-1, max_gen_seq_length)

        return GenericOutputs(
            sequences=outputs.sequences.long(),
            sequences_scores=outputs.sequences_scores,
        )


class QWen3GRPOLoraForGeneration(GenericPeftModel):
    prefix_keys_in_state_dict = {"^(?!model\.model\.).*": "model."}
    replace_keys_in_state_dict = {
        "q_proj.weight": "q_proj.base_layer.weight",
        "q_proj.bias": "q_proj.base_layer.bias",
        "v_proj.weight": "v_proj.base_layer.weight",
        "v_proj.bias": "v_proj.base_layer.bias",
    }

    def __init__(
        self,
        config_path: str,
        lora_r: Optional[int] = 16,
        lora_alpha: Optional[int] = 32,
        lora_dropout: Optional[float] = 0.05,
        fan_in_fan_out: Optional[bool] = True,
        target_modules: Optional[Union[List[str], str]] = ["q_proj", "v_proj"],
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Bloom Lora model for text generation tasks.

        Args:
            config_path (str): Path to the model configuration file.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__()
        self.config = Qwen3Config.from_json_file(config_path)
        self.config.gradient_checkpointing = gradient_checkpointing
        self.peft_config = LoraConfig(
            r=lora_r,
            lora_alpha=lora_alpha,
            lora_dropout=lora_dropout,
            fan_in_fan_out=fan_in_fan_out,
            target_modules=target_modules,
        )
        self.model = Qwen3ForCausalLM(self.config)
        self.model.add_adapter(self.peft_config)
        self.init_weights()

    def forward(
        self,
        input_ids: torch.Tensor,
        sampled_ids: torch.Tensor,
        sampled_rewards: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        sampled_attention_mask: Optional[torch.Tensor] = None,
    ):
        """
        Forward pass of the generation model.

        Args:
            input_ids (torch.Tensor, optional): Input tensor of shape (batch_size, sequence_length). Defaults to None.
            attention_mask (torch.Tensor, optional): Attention mask tensor of shape (batch_size, sequence_length). Defaults to None.
            position_ids (torch.Tensor, optional): Position IDs tensor of shape (batch_size, sequence_length). Defaults to None.

        Returns:
            torch Output logits.Tensor: tensor of shape (batch_size, sequence_length, vocab_size).
        """
        mean_rewards = sampled_rewards.mean(dim=-1)
        std_rewards = sampled_rewards.std(dim=-1)
        adv_rewards = sampled_rewards - mean_rewards.unsqueeze(-1)
        adv_rewards = adv_rewards / (std_rewards.unsqueeze(-1) + 1e-5)
        adv_rewards = adv_rewards.detach().reshape(-1)

        # 重复 input_ids 和 attention_mask
        G = sampled_ids.size(1)
        repeated_input_ids = input_ids.repeat_interleave(G, dim=0)  # [B*G, L_in]
        repeated_attention = (
            attention_mask.repeat_interleave(G, dim=0)
            if attention_mask is not None
            else None
        )
        sampled_seq_length = sampled_ids.size(-1)
        all_input_ids = torch.cat(
            [repeated_input_ids, sampled_ids.view(-1, sampled_seq_length)], dim=1
        )  # [B*G, L_in + L_out]

        if repeated_attention is not None:
            all_attention_mask = torch.cat(
                [
                    repeated_attention,
                    sampled_attention_mask.view(-1, sampled_seq_length),
                ],
                dim=1,
            )
        else:
            all_attention_mask = None

        all_outputs = self.model(
            input_ids=all_input_ids,
            attention_mask=all_attention_mask,
            return_dict=True,
        )

        input_seq_length = input_ids.size(1)
        logits = all_outputs.logits[:, input_seq_length - 1 : -1, :]

        labels = all_input_ids[:, input_seq_length:]
        labels_mask = all_attention_mask[:, input_seq_length:]
        nll_loss = F.cross_entropy(
            logits.reshape(-1, logits.size(-1)),
            labels.reshape(-1),
            reduction="none",
        ).reshape(labels.size(0), -1)
        logprobs = -nll_loss * labels_mask

        with torch.no_grad():
            self.model.disable_adapters()
            ref_outputs = self.model(
                input_ids=all_input_ids,
                attention_mask=all_attention_mask,
                return_dict=True,
            )
            ref_logits = ref_outputs.logits[:, input_seq_length - 1 : -1, :]
            ref_nll_loss = F.cross_entropy(
                ref_logits.reshape(-1, ref_logits.size(-1)),
                labels.reshape(-1),
                reduction="none",
            ).reshape(labels.size(0), -1)
            ref_logprobs = -ref_nll_loss * labels_mask
            self.model.enable_adapters()

        logprob_mean = logprobs.sum(-1) / (labels_mask.sum(-1) + 1e-5)
        ref_logprob_mean = ref_logprobs.sum(-1) / (labels_mask.sum(-1) + 1e-5)

        log_ratio = logprob_mean - ref_logprob_mean.detach()
        kl_div = 0.02 * (log_ratio**2)
        loss = (-adv_rewards.reshape(-1) * log_ratio + kl_div).mean()

        return loss

    @torch.no_grad()
    def generate(
        self,
        input_ids: torch.Tensor,
        num_beams: Optional[int] = 5,
        decoder_start_token_id: Optional[int] = 151643,
        decoder_end_token_id: Optional[Union[int, List[int]]] = 151645,
        decoder_pad_token_id: Optional[int] = 151643,
        num_return_sequences: Optional[int] = 1,
        min_gen_seq_length: Optional[int] = 0,
        max_gen_seq_length: Optional[int] = 512,
        repetition_penalty: Optional[float] = 1.0,
        no_repeat_ngram_size: Optional[int] = 0,
        early_stopping: Optional[bool] = True,
        length_penalty: Optional[float] = 1.0,
        num_beam_groups: Optional[int] = 1,
        diversity_penalty: Optional[float] = 0.0,
        do_sample: Optional[bool] = False,
        temperature: Optional[float] = 1.0,
        top_k: Optional[int] = 50,
        top_p: Optional[float] = 1.0,
    ):
        """
        Generate text using the generation model.

        Args:
            input_ids: Input tensor of shape (batch_size, sequence_length).
            num_beams (int, optional): Number of beams for beam search. Defaults to 5.
            decoder_start_token_id (int, optional): The ID of the decoder start token. Defaults to 2.
            decoder_end_token_id (int or List[int], optional): The ID(s) of the decoder end token(s). Defaults to 2.
            num_return_sequences (int, optional): Number of generated sequences to return. Defaults to 1.
            min_gen_seq_length (int, optional): Minimum length of generated sequences. Defaults to 0.
            max_gen_seq_length (int, optional): Maximum length of generated sequences. Defaults to 48.
            repetition_penalty (float, optional): Penalty for repeated tokens. Defaults to 1.0.
            no_repeat_ngram_size (int, optional): Size of n-grams to avoid repeating. Defaults to 0.
            early_stopping (bool, optional): Whether to stop generation early. Defaults to True.
            length_penalty (float, optional): Penalty for longer sequences. Defaults to 1.0.
            num_beam_groups (int, optional): Number of beam groups for diverse beam search. Defaults to 1.
            diversity_penalty (float, optional): Penalty for diverse sequences in diverse beam search. Defaults to 0.0.
            do_sample (bool, optional): Whether to use sampling for generation. Defaults to False.
            temperature (float, optional): Sampling temperature. Defaults to 1.0.
            top_k (int, optional): Top-k value for sampling. Defaults to 50.
            top_p (float, optional): Top-p value for sampling. Defaults to 1.0.

        Returns:
            GenericOutputs: Generated sequences and their scores.
        """
        input_seq_length = input_ids.size(1)
        outputs = self.model.generate(
            input_ids=input_ids,
            max_length=max_gen_seq_length + input_seq_length,
            min_length=min_gen_seq_length + input_seq_length,
            num_beams=num_beams,
            do_sample=do_sample,
            no_repeat_ngram_size=no_repeat_ngram_size,
            early_stopping=early_stopping,
            length_penalty=length_penalty,
            repetition_penalty=repetition_penalty,
            num_return_sequences=num_return_sequences,
            bos_token_id=decoder_start_token_id,
            eos_token_id=decoder_end_token_id,
            pad_token_id=decoder_pad_token_id,
            num_beam_groups=num_beam_groups,
            diversity_penalty=diversity_penalty,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            return_dict_in_generate=True,
            output_scores=True,
        )

        sequences = outputs.sequences.reshape(
            -1, num_return_sequences, outputs.sequences.size(-1)
        )
        outputs.sequences = (
            torch.zeros(sequences.size(0), num_return_sequences, max_gen_seq_length).to(
                device=sequences.device
            )
            + decoder_start_token_id
        )
        outputs.sequences[:, :, : sequences.size(-1) - input_seq_length].copy_(
            sequences[:, :, input_seq_length : sequences.size(-1)]
        )

        if num_return_sequences == 1:
            outputs.sequences = outputs.sequences.reshape(-1, max_gen_seq_length)

        return GenericOutputs(
            sequences=outputs.sequences.long(),
            sequences_scores=outputs.sequences_scores,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import json
import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from peft import LoraConfig, PeftModelForCausalLM
from transformers.models.qwen2_5_vl import (
    Qwen2_5_VLConfig,
    Qwen2_5_VLModel,
    Qwen2_5_VLForConditionalGeneration,
)
from unitorch.models import GenericModel, GenericOutputs
from unitorch.models.peft import PeftModelForSequenceClassification, GenericPeftModel


class QWen2_5VLLoraForGeneration(GenericPeftModel):
    prefix_keys_in_state_dict = {
        "^visual.": "model.model.",
        "^model(?!\.model).": "model.model.language_",
    }
    replace_keys_in_state_dict = {
        "q_proj.weight": "q_proj.base_layer.weight",
        "q_proj.bias": "q_proj.base_layer.bias",
        "v_proj.weight": "v_proj.base_layer.weight",
        "v_proj.bias": "v_proj.base_layer.bias",
    }

    def __init__(
        self,
        config_path: str,
        lora_r: Optional[int] = 16,
        lora_alpha: Optional[int] = 32,
        lora_dropout: Optional[float] = 0.05,
        fan_in_fan_out: Optional[bool] = True,
        target_modules: Optional[Union[List[str], str]] = ["q_proj", "v_proj"],
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Bloom Lora model for text generation tasks.

        Args:
            config_path (str): Path to the model configuration file.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__()
        self.config = Qwen2_5_VLConfig.from_json_file(config_path)
        self.config.gradient_checkpointing = gradient_checkpointing
        self.peft_config = LoraConfig(
            r=lora_r,
            lora_alpha=lora_alpha,
            lora_dropout=lora_dropout,
            fan_in_fan_out=fan_in_fan_out,
            target_modules=target_modules,
        )
        self.model = Qwen2_5_VLForConditionalGeneration(self.config)
        self.model.add_adapter(self.peft_config)
        self.init_weights()

    def forward(
        self,
        input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        image_grid_thw: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
    ):
        """
        Forward pass of the generation model.

        Args:
            input_ids (torch.Tensor, optional): Input tensor of shape (batch_size, sequence_length). Defaults to None.
            attention_mask (torch.Tensor, optional): Attention mask tensor of shape (batch_size, sequence_length). Defaults to None.
            position_ids (torch.Tensor, optional): Position IDs tensor of shape (batch_size, sequence_length). Defaults to None.

        Returns:
            torch Output logits.Tensor: tensor of shape (batch_size, sequence_length, vocab_size).
        """
        image_grid_thw = image_grid_thw.view(-1, image_grid_thw.size(-1))
        pixel_values = pixel_values.view(-1, pixel_values.size(-1))
        outputs = self.model(
            input_ids=input_ids,
            pixel_values=pixel_values,
            image_grid_thw=image_grid_thw,
            attention_mask=attention_mask,
            return_dict=True,
        )
        logits = outputs.logits
        return logits

    @torch.no_grad()
    def generate(
        self,
        input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        image_grid_thw: torch.Tensor,
        num_beams: Optional[int] = 5,
        decoder_start_token_id: Optional[int] = 151643,
        decoder_end_token_id: Optional[Union[int, List[int]]] = 151645,
        decoder_pad_token_id: Optional[int] = 151643,
        num_return_sequences: Optional[int] = 1,
        min_gen_seq_length: Optional[int] = 0,
        max_gen_seq_length: Optional[int] = 512,
        repetition_penalty: Optional[float] = 1.0,
        no_repeat_ngram_size: Optional[int] = 0,
        early_stopping: Optional[bool] = True,
        length_penalty: Optional[float] = 1.0,
        num_beam_groups: Optional[int] = 1,
        diversity_penalty: Optional[float] = 0.0,
        do_sample: Optional[bool] = False,
        temperature: Optional[float] = 1.0,
        top_k: Optional[int] = 50,
        top_p: Optional[float] = 1.0,
    ):
        """
        Generate text using the generation model.

        Args:
            input_ids: Input tensor of shape (batch_size, sequence_length).
            num_beams (int, optional): Number of beams for beam search. Defaults to 5.
            decoder_start_token_id (int, optional): The ID of the decoder start token. Defaults to 2.
            decoder_end_token_id (int or List[int], optional): The ID(s) of the decoder end token(s). Defaults to 2.
            num_return_sequences (int, optional): Number of generated sequences to return. Defaults to 1.
            min_gen_seq_length (int, optional): Minimum length of generated sequences. Defaults to 0.
            max_gen_seq_length (int, optional): Maximum length of generated sequences. Defaults to 48.
            repetition_penalty (float, optional): Penalty for repeated tokens. Defaults to 1.0.
            no_repeat_ngram_size (int, optional): Size of n-grams to avoid repeating. Defaults to 0.
            early_stopping (bool, optional): Whether to stop generation early. Defaults to True.
            length_penalty (float, optional): Penalty for longer sequences. Defaults to 1.0.
            num_beam_groups (int, optional): Number of beam groups for diverse beam search. Defaults to 1.
            diversity_penalty (float, optional): Penalty for diverse sequences in diverse beam search. Defaults to 0.0.
            do_sample (bool, optional): Whether to use sampling for generation. Defaults to False.
            temperature (float, optional): Sampling temperature. Defaults to 1.0.
            top_k (int, optional): Top-k value for sampling. Defaults to 50.
            top_p (float, optional): Top-p value for sampling. Defaults to 1.0.

        Returns:
            GenericOutputs: Generated sequences and their scores.
        """
        input_seq_length = input_ids.size(1)
        image_grid_thw = image_grid_thw.view(-1, image_grid_thw.size(-1))
        pixel_values = pixel_values.view(-1, pixel_values.size(-1))
        outputs = self.model.generate(
            input_ids=input_ids,
            pixel_values=pixel_values,
            image_grid_thw=image_grid_thw,
            max_length=max_gen_seq_length + input_seq_length,
            min_length=min_gen_seq_length + input_seq_length,
            num_beams=num_beams,
            do_sample=do_sample,
            no_repeat_ngram_size=no_repeat_ngram_size,
            early_stopping=early_stopping,
            length_penalty=length_penalty,
            repetition_penalty=repetition_penalty,
            num_return_sequences=num_return_sequences,
            bos_token_id=decoder_start_token_id,
            eos_token_id=decoder_end_token_id,
            pad_token_id=decoder_pad_token_id,
            num_beam_groups=num_beam_groups,
            diversity_penalty=diversity_penalty,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            return_dict_in_generate=True,
            output_scores=True,
        )

        sequences = outputs.sequences.reshape(
            -1, num_return_sequences, outputs.sequences.size(-1)
        )
        outputs.sequences = (
            torch.zeros(sequences.size(0), num_return_sequences, max_gen_seq_length).to(
                device=sequences.device
            )
            + decoder_start_token_id
        )
        outputs.sequences[:, :, : sequences.size(-1) - input_seq_length].copy_(
            sequences[:, :, input_seq_length : sequences.size(-1)]
        )

        if num_return_sequences == 1:
            outputs.sequences = outputs.sequences.reshape(-1, max_gen_seq_length)

        return GenericOutputs(
            sequences=outputs.sequences.long(),
            sequences_scores=outputs.sequences_scores,
        )


class QWen2_5VLDPOLoraForGeneration(GenericPeftModel):
    prefix_keys_in_state_dict = {
        "^visual.": "model.model.",
        "^model(?!\.model).": "model.model.language_",
    }
    replace_keys_in_state_dict = {
        "q_proj.weight": "q_proj.base_layer.weight",
        "q_proj.bias": "q_proj.base_layer.bias",
        "v_proj.weight": "v_proj.base_layer.weight",
        "v_proj.bias": "v_proj.base_layer.bias",
    }

    def __init__(
        self,
        config_path: str,
        lora_r: Optional[int] = 16,
        lora_alpha: Optional[int] = 32,
        lora_dropout: Optional[float] = 0.05,
        fan_in_fan_out: Optional[bool] = True,
        target_modules: Optional[Union[List[str], str]] = ["q_proj", "v_proj"],
        gradient_checkpointing: Optional[bool] = False,
        dpo_beta: Optional[float] = 0.1,
    ):
        """
        Bloom Lora model for text generation tasks.

        Args:
            config_path (str): Path to the model configuration file.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__()
        self.config = Qwen2_5_VLConfig.from_json_file(config_path)
        self.config.gradient_checkpointing = gradient_checkpointing
        self.peft_config = LoraConfig(
            r=lora_r,
            lora_alpha=lora_alpha,
            lora_dropout=lora_dropout,
            fan_in_fan_out=fan_in_fan_out,
            target_modules=target_modules,
        )
        self.model = Qwen2_5_VLForConditionalGeneration(self.config)
        self.model.add_adapter(self.peft_config)
        self.init_weights()
        self.dpo_beta = dpo_beta

    def forward(
        self,
        input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        image_grid_thw: torch.Tensor,
        win_input_ids: torch.Tensor,
        lose_input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        win_attention_mask: Optional[torch.Tensor] = None,
        lose_attention_mask: Optional[torch.Tensor] = None,
    ):
        """
        Forward pass of the generation model.

        Args:
            input_ids (torch.Tensor, optional): Input tensor of shape (batch_size, sequence_length). Defaults to None.
            attention_mask (torch.Tensor, optional): Attention mask tensor of shape (batch_size, sequence_length). Defaults to None.
            position_ids (torch.Tensor, optional): Position IDs tensor of shape (batch_size, sequence_length). Defaults to None.

        Returns:
            torch Output logits.Tensor: tensor of shape (batch_size, sequence_length, vocab_size).
        """
        image_grid_thw = image_grid_thw.view(-1, image_grid_thw.size(-1))
        pixel_values = pixel_values.view(-1, pixel_values.size(-1))
        win_input_ids = torch.cat([input_ids, win_input_ids], dim=1)
        lose_input_ids = torch.cat([input_ids, lose_input_ids], dim=1)
        if attention_mask is not None and win_attention_mask is not None:
            win_attention_mask = torch.cat([attention_mask, win_attention_mask], dim=1)
        if attention_mask is not None and lose_attention_mask is not None:
            lose_attention_mask = torch.cat(
                [attention_mask, lose_attention_mask], dim=1
            )
        win_outputs = self.model(
            input_ids=win_input_ids,
            pixel_values=pixel_values,
            image_grid_thw=image_grid_thw,
            attention_mask=win_attention_mask,
            return_dict=True,
        )
        lose_outputs = self.model(
            input_ids=lose_input_ids,
            pixel_values=pixel_values,
            image_grid_thw=image_grid_thw,
            attention_mask=lose_attention_mask,
            return_dict=True,
        )
        input_seq_length = input_ids.size(1)
        win_logits = win_outputs.logits[:, input_seq_length - 1 : -1, :]
        lose_logits = lose_outputs.logits[:, input_seq_length - 1 : -1, :]
        win_labels = win_input_ids[:, input_seq_length:]
        lose_labels = lose_input_ids[:, input_seq_length:]
        win_labels_mask = win_attention_mask[:, input_seq_length:]
        lose_labels_mask = lose_attention_mask[:, input_seq_length:]
        win_nll_loss = F.cross_entropy(
            win_logits.reshape(-1, win_logits.size(-1)),
            win_labels.reshape(-1),
            reduction="none",
        ).reshape(win_labels.size(0), -1)
        win_logprobs = -win_nll_loss * win_labels_mask
        lose_nll_loss = F.cross_entropy(
            lose_logits.reshape(-1, lose_logits.size(-1)),
            lose_labels.reshape(-1),
            reduction="none",
        ).reshape(lose_labels.size(0), -1)
        lose_logprobs = -lose_nll_loss * lose_labels_mask

        with torch.no_grad():
            self.model.disable_adapters()
            ref_win_outputs = self.model(
                input_ids=win_input_ids,
                attention_mask=win_attention_mask,
                return_dict=True,
            )
            ref_lose_outputs = self.model(
                input_ids=lose_input_ids,
                attention_mask=lose_attention_mask,
                return_dict=True,
            )
            ref_win_logits = ref_win_outputs.logits[:, input_seq_length - 1 : -1, :]
            ref_lose_logits = ref_lose_outputs.logits[:, input_seq_length - 1 : -1, :]
            ref_win_nll_loss = F.cross_entropy(
                ref_win_logits.reshape(-1, ref_win_logits.size(-1)),
                win_labels.reshape(-1),
                reduction="none",
            ).reshape(win_labels.size(0), -1)
            ref_win_logprobs = -ref_win_nll_loss * win_labels_mask
            ref_lose_nll_loss = F.cross_entropy(
                ref_lose_logits.reshape(-1, ref_lose_logits.size(-1)),
                lose_labels.reshape(-1),
                reduction="none",
            ).reshape(lose_labels.size(0), -1)
            ref_lose_logprobs = -ref_lose_nll_loss * lose_labels_mask
            self.model.enable_adapters()

        logratios = win_logprobs - lose_logprobs
        ref_logratios = ref_win_logprobs - ref_lose_logprobs
        logits = logratios - ref_logratios
        loss = -F.logsigmoid(self.dpo_beta * logits).mean()
        return loss

    @torch.no_grad()
    def generate(
        self,
        input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        image_grid_thw: torch.Tensor,
        num_beams: Optional[int] = 5,
        decoder_start_token_id: Optional[int] = 151643,
        decoder_end_token_id: Optional[Union[int, List[int]]] = 151645,
        decoder_pad_token_id: Optional[int] = 151643,
        num_return_sequences: Optional[int] = 1,
        min_gen_seq_length: Optional[int] = 0,
        max_gen_seq_length: Optional[int] = 512,
        repetition_penalty: Optional[float] = 1.0,
        no_repeat_ngram_size: Optional[int] = 0,
        early_stopping: Optional[bool] = True,
        length_penalty: Optional[float] = 1.0,
        num_beam_groups: Optional[int] = 1,
        diversity_penalty: Optional[float] = 0.0,
        do_sample: Optional[bool] = False,
        temperature: Optional[float] = 1.0,
        top_k: Optional[int] = 50,
        top_p: Optional[float] = 1.0,
    ):
        """
        Generate text using the generation model.

        Args:
            input_ids: Input tensor of shape (batch_size, sequence_length).
            num_beams (int, optional): Number of beams for beam search. Defaults to 5.
            decoder_start_token_id (int, optional): The ID of the decoder start token. Defaults to 2.
            decoder_end_token_id (int or List[int], optional): The ID(s) of the decoder end token(s). Defaults to 2.
            num_return_sequences (int, optional): Number of generated sequences to return. Defaults to 1.
            min_gen_seq_length (int, optional): Minimum length of generated sequences. Defaults to 0.
            max_gen_seq_length (int, optional): Maximum length of generated sequences. Defaults to 48.
            repetition_penalty (float, optional): Penalty for repeated tokens. Defaults to 1.0.
            no_repeat_ngram_size (int, optional): Size of n-grams to avoid repeating. Defaults to 0.
            early_stopping (bool, optional): Whether to stop generation early. Defaults to True.
            length_penalty (float, optional): Penalty for longer sequences. Defaults to 1.0.
            num_beam_groups (int, optional): Number of beam groups for diverse beam search. Defaults to 1.
            diversity_penalty (float, optional): Penalty for diverse sequences in diverse beam search. Defaults to 0.0.
            do_sample (bool, optional): Whether to use sampling for generation. Defaults to False.
            temperature (float, optional): Sampling temperature. Defaults to 1.0.
            top_k (int, optional): Top-k value for sampling. Defaults to 50.
            top_p (float, optional): Top-p value for sampling. Defaults to 1.0.

        Returns:
            GenericOutputs: Generated sequences and their scores.
        """
        input_seq_length = input_ids.size(1)
        image_grid_thw = image_grid_thw.view(-1, image_grid_thw.size(-1))
        pixel_values = pixel_values.view(-1, pixel_values.size(-1))
        outputs = self.model.generate(
            input_ids=input_ids,
            pixel_values=pixel_values,
            image_grid_thw=image_grid_thw,
            max_length=max_gen_seq_length + input_seq_length,
            min_length=min_gen_seq_length + input_seq_length,
            num_beams=num_beams,
            do_sample=do_sample,
            no_repeat_ngram_size=no_repeat_ngram_size,
            early_stopping=early_stopping,
            length_penalty=length_penalty,
            repetition_penalty=repetition_penalty,
            num_return_sequences=num_return_sequences,
            bos_token_id=decoder_start_token_id,
            eos_token_id=decoder_end_token_id,
            pad_token_id=decoder_pad_token_id,
            num_beam_groups=num_beam_groups,
            diversity_penalty=diversity_penalty,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            return_dict_in_generate=True,
            output_scores=True,
        )

        sequences = outputs.sequences.reshape(
            -1, num_return_sequences, outputs.sequences.size(-1)
        )
        outputs.sequences = (
            torch.zeros(sequences.size(0), num_return_sequences, max_gen_seq_length).to(
                device=sequences.device
            )
            + decoder_start_token_id
        )
        outputs.sequences[:, :, : sequences.size(-1) - input_seq_length].copy_(
            sequences[:, :, input_seq_length : sequences.size(-1)]
        )

        if num_return_sequences == 1:
            outputs.sequences = outputs.sequences.reshape(-1, max_gen_seq_length)

        return GenericOutputs(
            sequences=outputs.sequences.long(),
            sequences_scores=outputs.sequences_scores,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import numpy as np
from PIL import Image, ImageFilter
from typing import Any, Dict, List, Optional, Tuple, Union
from unitorch.utils import is_opencv_available
from unitorch.cli import CoreConfigureParser
import unitorch.cli.pipelines.qwen_image.text2image
import unitorch.cli.pipelines.qwen_image.image_editing



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import io
import re
import time
import socket
import requests
import tempfile
import hashlib
import logging
import subprocess
import pandas as pd
import gradio as gr
from PIL import Image
from collections import Counter, defaultdict
from torch.hub import download_url_to_file
from unitorch import get_temp_dir
from unitorch.cli import CoreConfigureParser
from unitorch.cli import register_webui
from unitorch.cli.webuis import (
    create_element,
    create_accordion,
    create_row,
    create_column,
    create_flex_layout,
    create_group,
    create_tab,
    create_tabs,
    create_blocks,
)
from unitorch.cli.webuis import SimpleWebUI

_js = """
() => {
    const shortcuts = (e) => {
        const event = document.all ? window.event : e;
        if(e.target.tagName.toLowerCase() == "body") {
            const code = e.key;
            if (code.toLowerCase() === "arrowright") {
                document.getElementById("ut-labeling-submit").click();
                document.activeElement.blur();
                window.focus();
            } else {
                const choices = document.getElementById("ut-labeling-choices").getElementsByTagName("label");
                if (/^[1-9]$/.test(code)) {
                    const index = parseInt(code, 10) - 1; // 数字键1对应choices[0]
                    if (index >= 0 && index < choices.length) {
                        choices[index].click();
                        document.activeElement.blur();
                        window.focus();
                    } else {
                        console.warn("Key index out of range:", index);
                    }
                }

            }
        }
    };
    document.addEventListener("keyup", shortcuts);
    console.log("Shortcut keys for labeling loaded.");
}
"""

_css = """
img {
    transform-origin: bottom;
    transform: scale(0.85) !important;
}

video {
    transform-origin: bottom;
    transform: scale(0.85) !important;
}
"""


@register_webui("core/webui/labeling")
class LabelingWebUI(SimpleWebUI):
    def __init__(
        self,
        config: CoreConfigureParser,
    ):
        self._config = config
        config.set_default_section("core/webui/labeling")
        data_file = config.getoption("data_file", None)
        result_file = config.getoption("result_file", None)
        force_to_relabel = config.getoption("force_to_relabel", False)
        names = config.getoption("names", "*")
        quoting = config.getoption("quoting", 3)
        if isinstance(names, str) and names == "*":
            names = None
        if isinstance(names, str):
            names = re.split(r"[,;]", names)
            names = [n.strip() for n in names]

        sep = config.getoption("sep", "\t")
        temp_folder = config.getoption("temp_folder", get_temp_dir())
        os.makedirs(temp_folder, exist_ok=True)
        self.temp_folder = temp_folder
        self.tags = config.getoption("tags", "#Labeling")

        self.dataset = pd.read_csv(
            data_file,
            names=names,
            header="infer" if names is None else None,
            sep=sep,
            quoting=quoting,
        )
        self.dataset["Index"] = self.dataset.index.map(lambda x: f"No.{x}")
        self.result_file = result_file

        self.http_url = "/gradio_api/file={0}"
        # show columns
        self.group_text_cols = config.getoption("group_text_cols", None)
        self.text_cols = config.getoption("text_cols", None)
        self.image_cols = config.getoption("image_cols", None)
        self.video_cols = config.getoption("video_cols", None)
        self.url_cols = config.getoption("url_cols", None)
        self.html_cols = config.getoption("html_cols", None)
        self.show_cols = config.getoption("show_cols", None)
        self.zip_cols = config.getoption("zip_cols", None)
        self.zip_http_url = config.getoption("zip_http_url", None)
        self.group_col = config.getoption("group_col", None)
        self.pre_label_col = config.getoption("pre_label_col", None)
        self.num_group_texts_per_row = config.getoption("num_group_texts_per_row", 4)
        self.num_images_per_row = config.getoption("num_images_per_row", 4)
        self.num_videos_per_row = config.getoption("num_videos_per_row", 4)
        self.num_mix_images_videos_per_row = config.getoption(
            "num_mix_images_videos_per_row", 4
        )
        self.num_html_per_row = config.getoption("num_html_per_row", 4)

        # css settings
        self.min_image_width = config.getoption("min_image_width", "none")
        self.min_image_height = config.getoption("min_image_height", "none")
        self.max_image_width = config.getoption("max_image_width", "none")
        self.max_image_height = config.getoption("max_image_height", "100px")
        self.min_video_width = config.getoption("min_video_width", "none")
        self.min_video_height = config.getoption("min_video_height", "none")
        self.max_video_width = config.getoption("max_video_width", "none")
        self.max_video_height = config.getoption("max_video_height", "100px")

        if self.group_text_cols is not None:
            if isinstance(self.group_text_cols, str):
                self.group_text_cols = re.split(r"[,;]", self.group_text_cols)
                self.group_text_cols = [n.strip() for n in self.group_text_cols]
            assert all(
                [col in self.dataset.columns for col in self.group_text_cols]
            ), f"group_text_cols {self.group_text_cols} not found in dataset"
        else:
            self.group_text_cols = []

        if self.text_cols is not None:
            if isinstance(self.text_cols, str):
                self.text_cols = re.split(r"[,;]", self.text_cols)
                self.text_cols = [n.strip() for n in self.text_cols]

            self.text_cols = [
                col for col in self.text_cols if col not in self.group_text_cols
            ] + self.group_text_cols
            assert all(
                [col in self.dataset.columns for col in self.text_cols]
            ), f"text_cols {self.text_cols} not found in dataset"
        else:
            self.text_cols = self.group_text_cols

        if self.image_cols is not None:
            if isinstance(self.image_cols, str):
                self.image_cols = re.split(r"[,;]", self.image_cols)
                self.image_cols = [n.strip() for n in self.image_cols]
            assert all(
                [col in self.dataset.columns for col in self.image_cols]
            ), f"image_cols {self.image_cols} not found in dataset"
        else:
            self.image_cols = []

        if self.video_cols is not None:
            if isinstance(self.video_cols, str):
                self.video_cols = re.split(r"[,;]", self.video_cols)
                self.video_cols = [n.strip() for n in self.video_cols]
            assert all(
                [col in self.dataset.columns for col in self.video_cols]
            ), f"video_cols {self.video_cols} not found in dataset"
        else:
            self.video_cols = []

        if self.url_cols is not None:
            if isinstance(self.url_cols, str):
                self.url_cols = re.split(r"[,;]", self.url_cols)
                self.url_cols = [n.strip() for n in self.url_cols]
            assert all(
                [col in self.dataset.columns for col in self.url_cols]
            ), f"url_cols {self.url_cols} not found in dataset"
        else:
            self.url_cols = []

        if self.html_cols is not None:
            if isinstance(self.html_cols, str):
                self.html_cols = re.split(r"[,;]", self.html_cols)
                self.html_cols = [n.strip() for n in self.html_cols]
            assert all(
                [col in self.dataset.columns for col in self.html_cols]
            ), f"html_cols {self.html_cols} not found in dataset"
        else:
            self.html_cols = []

        if self.show_cols is not None:
            if isinstance(self.show_cols, str):
                self.show_cols = re.split(r"[,;]", self.show_cols)
                self.show_cols = [n.strip() for n in self.show_cols]
            assert all(
                [col in self.dataset.columns for col in self.show_cols]
            ), f"show_cols {self.show_cols} not found in dataset"
        else:
            self.show_cols = list(self.dataset.columns)

        if self.zip_cols is not None:
            if isinstance(self.zip_cols, str):
                self.zip_cols = re.split(r"[,;]", self.zip_cols)
                self.zip_cols = [n.strip() for n in self.zip_cols]
            assert all(
                [col in self.dataset.columns for col in self.zip_cols]
            ), f"zip_cols {self.zip_cols} not found in dataset"
        else:
            self.zip_cols = []

        if len(self.zip_cols) > 0:
            assert self.zip_http_url is not None, "zip_http_url is required."

        self.num_group_text_cols = len(self.group_text_cols)
        self.num_text_cols = 0 if self.text_cols is None else len(self.text_cols)
        self.num_image_cols = 0 if self.image_cols is None else len(self.image_cols)
        self.num_video_cols = 0 if self.video_cols is None else len(self.video_cols)
        self.num_html_cols = 0 if self.html_cols is None else len(self.html_cols)
        self.guideline = config.getoption("guideline", None)
        self.choices = config.getoption("choices", None)
        self.checkbox = config.getoption("checkbox", False)
        self.use_shortcuts = config.getoption("use_shortcuts", False)
        self.html_styles = config.getoption("html_styles", {})
        self.dataset["User"] = ""
        self.dataset["Comment"] = ""
        self.dataset["Label"] = ""

        if self.pre_label_col is not None:
            assert (
                self.pre_label_col in self.dataset.columns
            ), f"pre_label_col {self.pre_label_col} not found in dataset"
            self.dataset["Label"] = self.dataset[self.pre_label_col].map(
                lambda x: str(x).strip()
            )
            self.dataset["Label"].fillna("", inplace=True)

        if isinstance(self.choices, str):
            self.choices = re.split(r"[,;]", self.choices)
            self.choices = [c.strip() for c in self.choices]
        self.choices = [str(c).strip() for c in self.choices]

        if os.path.exists(result_file) and not force_to_relabel:
            self.dataset = pd.read_csv(result_file, sep="\t")
        else:
            self.dataset.to_csv(result_file, sep="\t", index=False)

        self.logs = f"* {time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())}: Start Labeling. \n"

        # format dataset columns
        str_columns = ["User", "Comment", "Label"]
        str_columns += [self.group_col] if self.group_col is not None else []
        str_columns += (
            self.image_cols
            + self.video_cols
            + self.html_cols
            + self.url_cols
            + self.zip_cols
        )
        for col in str_columns:
            self.dataset[col].fillna("", inplace=True)
            self.dataset[col] = self.dataset[col].map(str)

        # create elements
        index = create_element(
            "dropdown",
            label="Index",
            default=" - ",
            values=[" - "] + self.dataset["Index"].tolist(),
        )
        if self.group_col is not None:
            group_values = self.dataset[self.group_col].unique().tolist()
        else:
            group_values = []
        group = create_element(
            "dropdown",
            label="Group",
            default=" - ",
            values=[" - "] + group_values,
        )
        user = create_element(
            "text",
            label="User",
            interactive=False,
            scale=2,
        )
        guideline_header = create_element(
            "markdown",
            label="# <div style='margin-top:10px'>Guideline</div>",
            interactive=False,
        )
        guideline = create_element(
            "markdown",
            label=f"{self.guideline}",
            interactive=False,
        )
        texts = [
            create_element(
                "text",
                label=col,
                lines=2,
            )
            for col in self.text_cols
        ]
        images = [
            create_element(
                "image",
                label=col,
            )
            for col in self.image_cols
        ]
        videos = [
            create_element(
                "video",
                label=col,
            )
            for col in self.video_cols
        ]
        htmls = [
            create_element(
                "html",
                label=col,
            )
            for col in self.html_cols
        ]
        choices = create_element(
            "radio" if not self.checkbox else "checkboxgroup",
            label="Label",
            values=self.choices,
            scale=3,
            elem_id="ut-labeling-choices",
        )

        comment = create_element(
            "text",
            label="Comment",
            lines=4,
            scale=2,
        )
        random_type = create_element(
            "radio",
            values=["All", "Labeled", "Unlabeled"],
            default="Unlabeled",
            label="Data Type",
            scale=2,
        )
        random = create_element(
            "button",
            label="Sample",
            scale=2,
        )
        submit = create_element(
            "button",
            label="Submit",
            elem_id="ut-labeling-submit",
        )
        reset = create_element(
            "button",
            label="Reset",
            variant="secondary",
        )
        progress = create_element(
            "text",
            label="Progress",
            interactive=False,
            scale=2,
        )
        res_disp_cols = create_element(
            "dropdown",
            label="Display Columns",
            default=" - ",
            values=[" - "] + self.show_cols,
            interactive=True,
            multiselect=True,
            scale=2,
        )
        res_label = create_element(
            "dropdown",
            label="Label",
            default=" - ",
            values=[" - "]
            + self.choices
            + ([] if not self.checkbox else [f"NOT {ch}" for ch in self.choices]),
        )

        adv_stats_header = create_element(
            "markdown",
            label="## <div style='margin-top:10px'>Distribution</div>",
            interactive=False,
        )
        adv_stats = create_element(
            "markdown",
            label="",
            interactive=False,
        )
        adv_logs_header = create_element(
            "markdown",
            label="## <div style='margin-top:10px'>Logs</div>",
            interactive=False,
        )
        adv_logs = create_element(
            "markdown",
            label="",
            interactive=False,
        )
        # show results
        data_type = create_element(
            "radio",
            values=["All", "Labeled", "Unlabeled"],
            default="Labeled",
            label="Data Type",
            scale=2,
        )
        refresh = create_element(
            "button",
            label="Refresh",
        )
        download = create_element(
            "download_button",
            label="Download",
        )
        results = create_element(
            "dataframe",
            label="Results",
            interactive=False,
        )

        # create blocks
        text_layout = (
            create_column(
                *texts[: -self.num_group_text_cols],
                create_flex_layout(
                    *texts[-self.num_group_text_cols :],
                    num_per_row=self.num_group_texts_per_row,
                ),
            )
            if self.num_group_text_cols > 0
            else create_column(*texts)
        )
        if (
            self.num_image_cols + self.num_video_cols
            > self.num_mix_images_videos_per_row
        ):
            image_layout = create_flex_layout(
                *images, num_per_row=self.num_images_per_row
            )
            video_layout = create_flex_layout(
                *videos, num_per_row=self.num_videos_per_row
            )
            mix_image_video_layout = None
        else:
            mix_image_video_layout = create_flex_layout(
                *images, *videos, num_per_row=self.num_mix_images_videos_per_row
            )
            image_layout, video_layout = None, None
        html_layout = create_flex_layout(*htmls, num_per_row=self.num_html_per_row)
        label_layout = create_row(
            comment,
            create_column(
                choices,
                create_row(reset, submit),
                scale=3,
            ),
        )

        layouts = []
        if self.num_text_cols > 0:
            layouts.append(text_layout)

        if self.num_image_cols > 0 and image_layout is not None:
            layouts.append(image_layout)

        if self.num_video_cols > 0 and video_layout is not None:
            layouts.append(video_layout)

        if mix_image_video_layout is not None:
            layouts.append(mix_image_video_layout)

        if self.num_html_cols > 0:
            layouts.append(html_layout)

        tab1 = create_tab(
            create_row(index, progress, user),
            create_row(group, random_type, random),
            *layouts,
            label_layout,
            name="Labeling",
        )
        tab2 = create_tab(
            create_row(progress, group, res_label, download),
            create_row(data_type, res_disp_cols, refresh),
            results,
            name="Results",
        )
        tab3 = create_tab(
            adv_stats_header,
            adv_stats,
            adv_logs_header,
            adv_logs,
            name="Advanced",
        )
        tabs = create_tabs(tab1, tab2, tab3)
        if self.use_shortcuts:
            iface = create_blocks(guideline_header, guideline, tabs, js=_js, css=_css)
        else:
            iface = create_blocks(guideline_header, guideline, tabs, css=_css)

        # create events
        iface.__enter__()
        submit.click(
            self.label,
            inputs=[index, group, user, choices, comment, adv_logs],
            outputs=[download, adv_stats, adv_logs, index],
            trigger_mode="once",
        )
        random.click(
            self.sample,
            inputs=[group, random_type],
            outputs=[index],
            trigger_mode="once",
        )
        refresh.click(
            self.show,
            inputs=[group, data_type, res_label, res_disp_cols],
            outputs=[progress, results],
            trigger_mode="once",
        )
        reset.click(
            self.reset,
            inputs=[index],
            outputs=[choices, comment, progress],
            trigger_mode="once",
        )
        index.change(
            fn=self.load,
            inputs=[index],
            outputs=[
                progress,
                group,
                choices,
                comment,
                *texts,
                *images,
                *videos,
                *htmls,
            ],
            trigger_mode="once",
        )

        iface.load(
            fn=self.sample,
            inputs=[group, random_type],
            outputs=[index],
        )
        iface.load(
            fn=lambda: tuple(self.show())
            + (os.path.abspath(self.result_file), self.stats()),
            inputs=[],
            outputs=[progress, results, download, adv_stats],
        )

        def get_user(request: gr.Request):
            if request:
                return request.username
            return None

        iface.load(fn=get_user, inputs=None, outputs=user)
        iface.load(fn=lambda: self.logs, inputs=None, outputs=adv_logs)

        iface.__exit__()

        super().__init__(config, iname="Labeling Tools", iface=iface)

    def process_sample(self, sample):
        def save_url(url):
            if url.startswith("http"):
                name = os.path.join(
                    self.temp_folder, hashlib.md5(url.encode()).hexdigest()
                )
                try:
                    if not os.path.exists(name):
                        download_url_to_file(url, name, progress=False)
                    return name
                except Exception as e:
                    print(e)
            return url

        for col in set(self.zip_cols):
            sample[col] = self.zip_http_url.format(sample[col])

        for col in self.image_cols + self.video_cols:
            sample[col] = save_url(sample[col])

        url = lambda x: (
            x if x.startswith("http") else self.http_url.format(os.path.abspath(x))
        )
        for col in self.url_cols:
            sample[col] = url(sample[col])

        for col in self.html_cols:
            sample[col] = self.html_styles.get(sample[col], "{0}").format(sample[col])
        return sample

    def process_results(self, results):
        url = lambda x: (
            x if x.startswith("http") else self.http_url.format(os.path.abspath(x))
        )

        for col in set(self.zip_cols):
            results[col] = results[col].map(lambda x: self.zip_http_url.format(x))

        for col in set(self.image_cols):
            results[col] = results[col].map(url)
            results[col] = results[col].map(
                lambda x: f'<img src="{x}" style="min-width: {self.min_image_width}; max-width: {self.max_image_width}; min-height: {self.min_image_height}; max-height: {self.max_image_height}; overflow: hidden;">'
            )
        for col in set(self.video_cols):
            results[col] = results[col].map(
                lambda x: (
                    x
                    if x.startswith("http")
                    else self.http_url.format(os.path.abspath(x))
                )
            )
            results[col] = results[col].map(
                lambda x: f'<video src="{x}" style="min-width: {self.min_video_width}; max-width: {self.max_video_width}; min-height: {self.min_video_height}; max-height: {self.max_video_height}; overflow: hidden;" preload="none" controls>'
            )

        for col in self.url_cols:
            results[col] = results[col].map(url)

        for col in set(self.html_cols):
            results[col] = results[col].map(
                lambda x: self.html_styles.get(x, "{0}").format(x)
            )
        return results

    def sample(self, group=None, random_type="Unlabeled"):
        if random_type == "Labeled":
            sampled_data = self.dataset[self.dataset["Label"] != ""]
        elif random_type == "Unlabeled":
            sampled_data = self.dataset[self.dataset["Label"] == ""]
        else:
            sampled_data = self.dataset

        if group is not None and group not in ["", " - "]:
            sampled_data = sampled_data[sampled_data[self.group_col] == group]

        if len(sampled_data) == 0:
            return " - "

        return sampled_data.sample(1).iloc[0]["Index"]

    def show(self, group=None, data_type="Labeled", choice=None, disp_cols=None):
        total = self.dataset.shape[0]
        progress = f"{len(self.dataset[self.dataset['Label'] != ''])} / {total}"

        results = self.dataset.copy()
        if data_type == "Labeled":
            results = results[results["Label"] != ""]
        elif data_type == "Unlabeled":
            results = results[results["Label"] == ""]

        if choice is not None and choice not in ["", " - "] and len(results) > 0:
            if self.checkbox and str(choice).startswith("NOT "):
                results = results[
                    results["Label"].map(lambda x: str(choice)[4:] not in x)
                ]
            else:
                results = results[results["Label"].map(lambda x: str(choice) in x)]

        if group is not None and group not in ["", " - "] and len(results) > 0:
            results = results[results[self.group_col] == str(group)]

        if len(results) > 0:
            results = self.process_results(results)

        if disp_cols is None:
            disp_cols = self.show_cols
        elif " - " in disp_cols:
            disp_cols = self.show_cols

        results = results[
            ["Index"]
            + [
                col
                for col in disp_cols
                if col not in ["Comment", "User", "Label", "Index"]
            ]
            + ["Label", "Comment"]
        ]
        return (progress, results)

    def reset(self, index):
        self.dataset.loc[self.dataset.Index == index, "User"] = ""
        self.dataset.loc[self.dataset.Index == index, "Label"] = ""
        self.dataset.loc[self.dataset.Index == index, "Comment"] = ""
        self.dataset.to_csv(self.result_file, sep="\t", index=False)
        total = self.dataset.shape[0]
        labeled = self.dataset[self.dataset["Label"] != ""].shape[0]
        progress = f"{labeled} / {total}"
        gr.Info(f"Reset {index} Success.")
        return None, None, progress

    def load(self, index):
        total = self.dataset.shape[0]
        progress = f"{len(self.dataset[self.dataset['Label'] != ''])} / {total}"

        if len(self.dataset[self.dataset["Index"] == index]) == 0:
            return (progress, None, None, None) + tuple(
                [None]
                * (
                    self.num_text_cols
                    + self.num_image_cols
                    + self.num_video_cols
                    + self.num_html_cols
                )
            )

        new_one = self.dataset[self.dataset["Index"] == index].iloc[0]
        new_one = self.process_sample(new_one)
        new_texts = new_one[self.text_cols].tolist()
        new_images = new_one[self.image_cols].tolist()
        new_videos = new_one[self.video_cols].tolist()
        new_htmls = new_one[self.html_cols].tolist()
        new_group = new_one[self.group_col] if self.group_col is not None else None
        new_comment = new_one["Comment"]
        new_choices = (
            new_one["Label"].split("<label-gap>") if new_one["Label"] != "" else None
        )
        new_choices = (
            new_choices[0]
            if not self.checkbox and new_choices is not None
            else new_choices
        )
        new_texts = [text if text != "" else None for text in new_texts]
        new_images = [img if img != "" else None for img in new_images]
        new_videos = [vid if vid != "" else None for vid in new_videos]
        new_htmls = [html if html != "" else None for html in new_htmls]

        return (
            (progress, new_group, new_choices, new_comment)
            + tuple(new_texts)
            + tuple(new_images)
            + tuple(new_videos)
            + tuple(new_htmls)
        )

    def stats(self):
        dataset = self.dataset.copy()
        dataset["Num"] = 1
        labeled = dataset[dataset["Label"] != ""]
        choices = self.choices

        labeled["Label"] = labeled["Label"].map(
            lambda x: x.split("<label-gap>") if x != "" else []
        )
        labeled_exploded = labeled.explode("Label")

        percent = lambda x, y: f"{x / (y if y > 0 else 1):.2%}"

        if self.group_col is not None:
            group = labeled.groupby(self.group_col)["Num"].count().to_dict()

            stats = labeled_exploded.groupby([self.group_col, "Label"])["Num"].count()
            stats = [
                {
                    "Group": g,
                    "Label": c,
                    "Count": f"{stats.get((g, c), 0)} / {group[g]}",
                    "Percentage": percent(stats.get((g, c), 0), group[g]),
                }
                for g in group.keys()
                for c in choices
            ]
            stats = pd.DataFrame(stats)
            stats = stats.append(
                {
                    "Group": "-",
                    "Label": "Total",
                    "Count": f"{len(labeled)} / {len(dataset)}",
                    "Percentage": percent(len(labeled), len(dataset)),
                },
                ignore_index=True,
            )

        else:
            stats = labeled_exploded.groupby("Label")["Num"].count()
            stats = [
                {
                    "Label": c,
                    "Count": stats.get(c, 0),
                    "Percentage": percent(stats.get(c, 0), len(labeled)),
                }
                for c in choices
            ]
            stats = pd.DataFrame(stats)
            stats = stats.append(
                {
                    "Label": "Total",
                    "Count": len(labeled),
                    "Percentage": percent(len(labeled), len(dataset)),
                },
                ignore_index=True,
            )

        stats = stats.to_markdown(index=False)
        return stats

    def label(
        self,
        index,
        group,
        user,
        choice,
        comment,
        logs,
    ):
        if (
            choice is None
            or choice == ""
            or (isinstance(choice, (list, tuple)) and len(choice) == 0)
        ):
            gr.Warning("Please ensure the label field is not left empty.")
            return os.path.abspath(self.result_file), self.stats(), logs, index
        if isinstance(choice, list) or isinstance(choice, tuple):
            choice = "<label-gap>".join(choice)
        self.dataset.loc[self.dataset.Index == index, "User"] = user
        self.dataset.loc[self.dataset.Index == index, "Label"] = choice
        self.dataset.loc[self.dataset.Index == index, "Comment"] = comment
        self.dataset.to_csv(self.result_file, sep="\t", index=False)

        if user is not None and user != "":
            new_logs = (
                f"* {time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())}: User {user} Label {index} to {choice} Success. \n"
                + self.logs
            )
        else:
            new_logs = (
                f"* {time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())}: Label {index} to {choice} Success. \n"
                + self.logs
            )
        self.logs = new_logs
        return (
            os.path.abspath(self.result_file),
            self.stats(),
            new_logs,
            self.sample(group),
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import unitorch.cli.fastapis.stable_flux.text2image
import unitorch.cli.fastapis.stable_flux.image2image
import unitorch.cli.fastapis.stable_flux.image_redux
import unitorch.cli.fastapis.stable_flux.inpainting
import unitorch.cli.fastapis.stable_flux.kontext2image
import unitorch.cli.fastapis.stable_flux.redux_inpainting
